/*
 * Copyright 2024 Luca Fulchir <luca.fulchir@runesauth.com>
 *
 * Licensed under the Apache License, Version 2.0 with LLVM exception (the
 * "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License and of the exception at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     https://spdx.org/licenses/LLVM-exception.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

use ::proc_macro::TokenStream;
use ::quote::quote;
use ::syn::{parse_macro_input, DeriveInput, Fields, ItemStruct};

pub(crate) fn repository(
    attrs: TokenStream,
    input: TokenStream,
) -> TokenStream {
    let base = parse_macro_input!(attrs as ItemStruct);
    let local = parse_macro_input!(input as ItemStruct);

    let Fields::Named(local_fields) = local.fields else {
        use ::syn::spanned::Spanned;
        return ::syn::Error::new(
            local.fields.span(),
            "unnamed fields are not supported",
        )
        .to_compile_error()
        .into();
    };

    let Fields::Named(base_fields) = base.fields else {
        use ::syn::spanned::Spanned;
        return ::syn::Error::new(
            base.fields.span(),
            "unnamed fields are not supported",
        )
        .to_compile_error()
        .into();
    };
    let local_fields_it = local_fields.named.iter();
    let base_fields_extra = base_fields.named.iter();
    let attrs = local.attrs;
    let generics = local.generics;
    let ident = local.ident;
    let vis = local.vis;
    quote! {
        #(#attrs)
        *
        #[::macro_magic::export_tokens]
        #[derive(::bok::Repository, Debug)]
        #vis struct #ident<#generics> {
            #(#base_fields_extra),
            *
            ,
            #(#local_fields_it),
            *
        }
    }
    .into()
}

pub(crate) fn derive_repository(input: TokenStream) -> TokenStream {
    let input = parse_macro_input!(input as DeriveInput);

    let name = input.ident.clone();

    let expanded = quote! {
        impl ::bok::Repository for #name {
            fn name(&self) -> ::bok::RepoName {
                (module_path!().to_owned() + "::" + ::std::stringify!(#name)).as_str().into()
            }
            fn path(&self) -> ::bok::Path<::bok::RepoName> {
                (module_path!().to_owned() + "::" + ::std::stringify!(#name)).as_str().into()
            }
        }
    };

    TokenStream::from(expanded)
}
